@@ -245,11 +245,82 @@ describe('Binding dependencies', function() {
         expect(testNode).toContainText('substitute');
         expect(observable.getSubscriptionsCount()).toEqual(1);
 
-        // uptdate observable to update binding
+        // update observable to update binding
         observable('new value');
         expect(testNode).toContainText('new value');
     });
 
+    it('Should not cause updates if an observable accessed in a childrenComplete callback is changed', function () {
+        ko.bindingHandlers.test = {
+            init: function() {
+                return { controlsDescendantBindings: true };
+            },
+            update: function(element, valueAccessor, allBindings, viewModel, bindingContext) {
+                ko.utils.unwrapObservable(valueAccessor());
+                element.innerHTML = "<span data-bind='text: childprop'></span>";
+                ko.applyBindingsToDescendants(bindingContext, element);
+            }
+        };
+
+        var callbackObservable = ko.observable(1),
+            bindingObservable = ko.observable(1),
+            callbacks = 0,
+            vm = {
+                childprop: 'child',
+                bindingObservable: bindingObservable,
+                callback: function () { callbackObservable(); callbacks++; }
+            };
+
+        testNode.innerHTML = "<div data-bind='test: bindingObservable, childrenComplete: callback'></div>";
+        ko.applyBindings(vm, testNode);
+        expect(callbacks).toEqual(1);
+
+        // Change the childprop which is not an observable so should not change the bound element
+        vm.childprop = 'new child';
+        expect(testNode.childNodes[0]).toContainText('child');
+
+        // Update callback observable and check that the binding wasn't updated
+        callbackObservable(2);
+        expect(testNode.childNodes[0]).toContainText('child');
+
+        // Update the bound observable and verify that the binding is now updated
+        bindingObservable(2);
+        expect(testNode.childNodes[0]).toContainText('new child');
+        expect(callbacks).toEqual(2);
+    });
+
+    it('Should always use the latest value of a childrenComplete callback', function () {
+        ko.bindingHandlers.test = {
+            init: function() {
+                return { controlsDescendantBindings: true };
+            },
+            update: function(element, valueAccessor, allBindings, viewModel, bindingContext) {
+                var innerContext = bindingContext.createChildContext({childprop: ko.utils.unwrapObservable(valueAccessor())});
+                element.innerHTML = "<span data-bind='text: childprop'></span>";
+                ko.applyBindingsToDescendants(innerContext, element);
+            }
+        };
+
+        var callbackSpy1 = jasmine.createSpy('callbackSpy1'),
+            callbackSpy2 = jasmine.createSpy('callbackSpy2'),
+            vm = {
+                observable: ko.observable('value'),
+                callback: callbackSpy1
+            };
+
+        testNode.innerHTML = "<div data-bind='test: observable, childrenComplete: callback'></div>";
+        ko.applyBindings(vm, testNode);
+        expect(callbackSpy1).toHaveBeenCalled();
+
+        callbackSpy1.reset();
+        vm.callback = callbackSpy2;
+
+        vm.observable('new value');
+        expect(testNode.childNodes[0]).toContainText('new value');
+        expect(callbackSpy1).not.toHaveBeenCalled();
+        expect(callbackSpy2).toHaveBeenCalled();
+    });
+
     describe('Observable view models', function() {
         it('Should update bindings (including callbacks)', function() {
             var vm = ko.observable(), clickedVM;
@@ -489,7 +560,7 @@ describe('Binding dependencies', function() {
         });
 
         it('Should leave bindings without an "after" value where they are', function() {
-            // This test is set up to be unambiguous, because only test1 and test2 are reorderd
+            // This test is set up to be unambiguous, because only test1 and test2 are reordered
             // (they have the dependency between them) and test3 is left alone.
             ko.bindingHandlers.test2.after = ['test1'];
             testNode.innerHTML = "<div data-bind='test2, test1, test3'></div>";
